;################################################################################
;#										#
;# chipbasic - single chip basic computer with ATmega32				#
;# full screen editor								#
;# copyright (c) 2006/2007 Joerg Wolfram (joerg@jcwolfram.de)			#
;#										#
;#										#
;# This program is free software; you can redistribute it and/or		#
;# modify it under the terms of the GNU General Public License			#
;# as published by the Free Software Foundation; either version 3		#
;# of the License, or (at your option) any later version.			#
;#										#
;# This program is distributed in the hope that it will be useful,		#
;# but WITHOUT ANY WARRANTY; without even the implied warranty of		#
;# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the GNU		#
;# General Public License for more details.					#
;#										#
;# You should have received a copy of the GNU General Public			#
;# License along with this library; if not, write to the			#
;# Free Software Foundation, Inc., 59 Temple Place - Suite 330,			#
;# Boston, MA 02111-1307, USA.							#
;#										#
;################################################################################

.set	editor_linestart = 2
.set	editor_linedigits = 2
.set	bas_linelen = 27
.set	editor_unchanged = 32
.set	editor_changed = '*'

;------------------------------------------------------------------------------
;main screen mask
;------------------------------------------------------------------------------
editor_entry:	push	ereg
		lds	tempreg1,libmio_ram+46	;prg to edit
		call	mem_setprog
		libmio_clrscr			;clear screen
		libmio_emode			;set screen in editor mode
		clr	ctrl			;set mode

		libmio_thistext
		.db 0,0,editor_unchanged,"PROGRAM: ",0
		ldi	YL,LOW(libmio_vram+9)	;start of program name
		ldi	YH,HIGH(libmio_vram+9)
		call	mem_readname
		rcall	editor_cstat		;clear status line
		ldi	XL,7
		sts	libmio_color,XL
		ldi	YL,0
		ldi	YH,2
		ldi	ZL,61
		ldi	ZH,2
		libmio_fbox
		call	lce_pload		;load program	
		pop	ereg
		cpi	ereg,0
		breq	editor_entry_p
editor_lnum_2:	ldi	XL,0xef			;yellow
		ldi	ZL,LOW(errtab1*2)
		ldi	ZH,HIGH(errtab1*2)
editor_lnum_3:	cpi	ereg,1
		breq	editor_entry_o
		ldi	XL,0xaf			;red
		dec	ereg
editor_lnum_4:	lpm	tempreg1,Z+
		cpi	tempreg1,0
		brne	editor_lnum_4
		rjmp	editor_lnum_3	

editor_entry_o:	rcall	editor_fstat		;fill status line
		ldi	ctrl,0x02
		ldi	XL,2
		ldi	XH,2
		libmio_gotoxy
		libmio_romtext			;errortext
		libmio_thistext
		.db	255," IN ",0
		clr	XH
		mov	XL,r12			;line no	
		inc	XL
    		libmio_outdez
		ldi	tempreg1,':'
		libmio_outchar
		mov	XL,r13			;statement	
		inc	XL
    		libmio_outdez
editor_entry_p:	ret				;thats all
	
editor_cset:	ldi	XL,editor_changed		;change signal char
		mov	tempreg6,XL
		ldi	tempreg4,editor_unchanged
		ret
	
;------------------------------------------------------------------------------
;main editor loop
;------------------------------------------------------------------------------
editor_main:	rcall	editor_cset
		rcall	editor_entry
		mov	r13,byte0		;first viewed line
editor_main_00:	clr	XL			;set cursor position
		clr	XH
editor_main_0:	sts	libmio_edstart,r13
editor_main_1:	libmio_ecursor

;------------------------------------------------------------------------------
;pos1 key
;------------------------------------------------------------------------------
		cpi	tempreg1,0xe0		;pos1
		brne	editor_main_2
		clr	XL			;X=0
		rjmp	editor_main_0

;------------------------------------------------------------------------------
;end key
;------------------------------------------------------------------------------
editor_main_2:	cpi	tempreg1,0xe1		;end
		brne	editor_main_3
		ldi	XL,libmio_ecols-4
		rjmp	editor_checkb

;------------------------------------------------------------------------------
;cursor left key
;------------------------------------------------------------------------------
editor_main_3:	cpi	tempreg1,0xe2		;left
		brne	editor_main_4
		dec	XL
		rjmp	editor_checkb

;------------------------------------------------------------------------------
;cursor right key
;------------------------------------------------------------------------------
editor_main_4:	cpi	tempreg1,0xe3		;right
		brne	editor_main_5
		inc	XL
		mov	ctrl,XH
		call	lce_fend		;find end
		cp	ctrl,XL
		brcc	editor_main_41
		cpi	XH,bas_lines-1
		brne	editor_main_40
		dec	XL
		rjmp	editor_checkb	
editor_main_40:	clr	XL
		inc	XH	
editor_main_41:	rjmp	editor_checkb

;------------------------------------------------------------------------------
;cursor up key
;------------------------------------------------------------------------------
editor_main_5:	cpi	tempreg1,0xe4		;up
		brne	editor_main5
		dec	XH
		sbrc	XH,7
		clr	XH
		rjmp	editor_checkb

;------------------------------------------------------------------------------
;page up key
;------------------------------------------------------------------------------
editor_main5:	cpi	tempreg1,0xe6		;page up
		brne	editor_main6
		subi	XH,0x08
		sbrc	XH,7
		clr	XH
		rjmp	editor_checkb

;------------------------------------------------------------------------------
;page down key
;------------------------------------------------------------------------------
editor_main6:	cpi	tempreg1,0xe7		;page down
		brne	editor_main_6
		subi	XH,0xf8
		rjmp	editor_checkb

;------------------------------------------------------------------------------
;cursor down key
;------------------------------------------------------------------------------
editor_main_6:	cpi	tempreg1,0xe5		;down
		brne	editor_main_7
		inc	XH
		rjmp	editor_checkb

;------------------------------------------------------------------------------
;enter key
;------------------------------------------------------------------------------
editor_main_7:	cpi	tempreg1,0xea		;enter
		brne	editor_main_70
		inc	XH
		clr	XL

;------------------------------------------------------------------------------
;limit coordinates to editable area
;------------------------------------------------------------------------------
editor_checkb:	cpi	XL,0x80			;out of left bound?
		brcs	editor_checkb1		;no
		ldi	XL,libmio_ecols-4	;set to right pos
		dec	XH			;prev line
		cpi	XH,0x80			;out of upper bound?
		brcs	editor_chb6		;no -> find line end
		clr	XL			;set pos to 0,0
		clr	XH
		rjmp	editor_main_0		;editor loop

editor_checkb1:	cpi	XL,libmio_ecols-3	;out of right bound?
		brcs	editor_chb3		;no
		cpi	XH,bas_lines-1		;last line?
		breq	editor_chb6		;yes -> find line end
		clr	XL			;X=0
		inc	XH			;next line
		rjmp	editor_main_0		;editor loop
		
editor_chb3:	cpi	XH,bas_lines
		brcs	editor_chb4
		ldi	XH,bas_lines-1		
		
editor_chb4:	mov	r16,XH			;y-coord
		cp	r16,r13			;check if line is visible
		brcc	editor_chb5
		mov	r13,r16
editor_chb5:	mov	r16,r13			;display Y
		subi	r16,0xf1		;+15
		cp	r16,XH			;check if line is visible
		brcc	editor_chb6
		mov	r16,XH
		subi	r16,0x0f
		mov	r13,r16
editor_chb6:	mov	ctrl,XH
		call	lce_fend		;find end
		cp	ctrl,XL
		brcc	editor_chb7
		mov	XL,ctrl	
editor_chb7:	rjmp	editor_main_0			

;------------------------------------------------------------------------------
;alt+delete key
;------------------------------------------------------------------------------
editor_main_70:	lds	tempreg2,libmio_kbdstate
		sbrs	tempreg2,4
		rjmp	editor_main_8
		mov	ctrl,XH
		cpi	tempreg1,0xe9		;delete
		brne	editor_main_75
		push	XL
		push	XH
		call	lce_del
		pop	XH
		pop	XL		
		rjmp	editor_change

;------------------------------------------------------------------------------
;alt+insert key (insert line)
;------------------------------------------------------------------------------
editor_main_75:	cpi	tempreg1,0xe8		;insert
		brne	editor_main_8
		push	XL
		push	XH
		call	lce_ins
		pop	XH
		pop	XL		
		rjmp	editor_change
		
;------------------------------------------------------------------------------
;delete key
;------------------------------------------------------------------------------
editor_main_8:	cpi	tempreg1,0xe9		;delete
		brne	editor_main_9
		rcall	editor_dchar
		rjmp	editor_change

;------------------------------------------------------------------------------
;backspace key
;------------------------------------------------------------------------------
editor_main_9:	cpi	tempreg1,0xec		;backspace
		brne	editor_main_a
		cpi	XL,0			;at the begin of the line?
		brne	editor_main_91		;no
		rjmp	editor_main_1		;loop
editor_main_91:	dec	XL
		rcall	editor_dchar
		rjmp	editor_change		;loop

;------------------------------------------------------------------------------
;F2 save
;------------------------------------------------------------------------------
editor_main_a:	cpi	tempreg1,0xf2
		brne	editor_main_c
		rcall	editor_uc
		breq	editor_main_b1
		ldi	XL,0xef
		rcall	editor_fstat		;clear status line
		libmio_thistext
		.db 2,2,"SAVE PROGRAM Y/N ?",0
		rcall	editor_yn		
		brtc	editor_main_b
		call	lce_psave
		call	lce_pload
		rcall	editor_cset
		sts	libmio_vram,tempreg4	;set unchange symbol
editor_main_b:	rcall	editor_cstat
		clr	ereg
editor_main_b1:	rjmp	editor_main_00			

;------------------------------------------------------------------------------
;F3 Load
;------------------------------------------------------------------------------
editor_main_c:	cpi	tempreg1,0xf3
		brne	editor_main_d
		rcall	editor_uc
		breq	editor_main_c1
		ldi	XL,0xef
		rcall	editor_fstat		;clear status line
		libmio_thistext
		.db 2,2,"LOST CHANGES Y/N ?",0
		rcall	editor_yn		
		brtc	editor_main_b
		call	lce_pload
		rcall	editor_cset
		sts	libmio_vram,tempreg4	;set unchange symbol
editor_main_c1:	rjmp	editor_main_b

;------------------------------------------------------------------------------
;F8 new program
;------------------------------------------------------------------------------
editor_main_d:	cpi	tempreg1,0xf8
		brne	editor_mainr
		ldi	XL,0xef
		rcall	editor_fstat		;clear status line
		libmio_thistext
		.db 2,2,"CLEAR PROGRAM Y/N ?",0
		rcall	editor_yn		
		brtc	editor_main_b
		call	lce_pnew
		rcall	editor_cset
		sts	libmio_vram,tempreg6	;set change symbol
editor_main_e:	rjmp	editor_main_b

;------------------------------------------------------------------------------
;F9 edit programm name
;------------------------------------------------------------------------------
editor_mainr:	cpi	tempreg1,0xf9
		brne	editor_copy
		mov	tempreg7,XL
		mov	tempreg8,XH
		rcall	editor_ren
		mov	XL,tempreg7
		mov	XH,tempreg8
		sts	libmio_vram,tempreg6	;set change symbol
		rjmp	editor_main_1

;------------------------------------------------------------------------------
;ctrl/c copy
;------------------------------------------------------------------------------
editor_copy:	sbrs	tempreg2,2
		rjmp	editor_exit
		cpi	tempreg1,'c'
		brne	editor_move
		ldi	tempreg3,'#'
		sts	libmio_vram+27,tempreg3	;set buffer symbol
		push	XL
		push	XH
		mov	ctrl,XH
		call	lce_rline		;copy to inbuf
editor_copy_1:	lds	tempreg3,bas_inbuf
		sts	libmio_vram+28,tempreg3	;copy number
		lds	tempreg3,bas_inbuf+1
		sts	libmio_vram+29,tempreg3	;copy number
		pop	XH
		pop	XL		
		rjmp	editor_main_1

;------------------------------------------------------------------------------
;ctrl/x move
;------------------------------------------------------------------------------
editor_move:	cpi	tempreg1,'x'
		brne	editor_paste
		ldi	tempreg3,'#'
		sts	libmio_vram+27,tempreg3	;set buffer symbol
		push	XL
		push	XH
		mov	ctrl,XH
		call	lce_mline		;move to inbuf
		sts	libmio_vram,tempreg6	;set change symbol
		rjmp	editor_copy_1

;------------------------------------------------------------------------------
;ctrl/v paste
;------------------------------------------------------------------------------
editor_paste:	cpi	tempreg1,'v'
		brne	editor_exit
		ldi	tempreg3,' '
		sts	libmio_vram+27,tempreg3	;clear copy symbol
		sts	libmio_vram+28,tempreg3	;clear number
		sts	libmio_vram+29,tempreg3	;clear number
		push	XL
		push	XH
		mov	ctrl,XH
		call	lce_wline		;copy from inbuf
		call	lce_renumber
		pop	XH
		pop	XL		
		sts	libmio_vram,tempreg6	;set change symbol
		rjmp	editor_main_1
		
editor_main_ex:	rjmp	editor_main_e		;jump extender

;------------------------------------------------------------------------------
;ESC
;------------------------------------------------------------------------------
editor_exit:	cpi	tempreg1,0xed
		brne	editor_main_k
		rcall	editor_uc
		brne	editor_exit_1
		libmio_nmode			;switch to normal mode
		ret
editor_exit_1:	ldi	XL,0xef
		rcall	editor_fstat
		libmio_thistext
		.db 2,2,"QUIT W/O SAVE Y/N ?",0
		rcall	editor_yn		
		brtc	editor_main_ex
editor_exit_2:	lds	XL,libmio_keycode
		cpi	XL,0
		brne	editor_exit_2		
		libmio_nmode			;switch to normal mode
		ret

;------------------------------------------------------------------------------
;F10 run
;------------------------------------------------------------------------------
editor_main_k:	cpi	tempreg1,0xfa		;F10
		brne	editor_main_f		;no
		rcall	editor_uc
		brne	editor_main_l
		libmio_clrscr			;clear screen
		libmio_nmode			;switch to normal mode
		ret				;go back and run
editor_main_l:	ldi	XL,0xbf			;magenta
		rcall	editor_fstat
		libmio_thistext
		.db 2,2,"SAVE CHANGES BEFORE RUN!",0
		libmio_waitkey		
		rjmp	editor_main_b

editor_main_f:
;------------------------------------------------------------------------------
;a visible  char
;------------------------------------------------------------------------------
editor_main_20:	lds	tempreg2,libmio_kbdstate
		sbrc	tempreg2,2		;left ctrl
		rjmp	editor_main_ig		;ignore
		sbrc	tempreg2,3		;right ctrl
		rjmp	editor_main_ig		;ignore
		cpi	tempreg1,0x1c		;degree
		brcs	editor_main_ig		;ignore less
		cpi	tempreg1,0x80		;last visible+1
		brcc	editor_main_ig		;igore greater or equal
		
		cpi	XL,libmio_ecols-4	;we are at the end?
		breq	editor_main_28		;yes
		
		mov	tempreg2,XL		;save X-pos
		ldi	XL,libmio_ecols-4
		libmio_ecalc
		
editor_main_21:	ld	tempreg3,-Y		;char on cursor pos		
		std	Y+1,tempreg3		;write to next
		dec	XL
		cp	XL,tempreg2
		brne	editor_main_21
    		st	Y,tempreg1
		inc	XL
		sts	libmio_vram,tempreg6	;change symbol
		rjmp	editor_main_1		
				
editor_main_28:	libmio_ecalc			;calculate vram address
    		st	Y,tempreg1
editor_change:	sts	libmio_vram,tempreg6	;set change symbol
		rjmp	editor_main_1		;loop
		
editor_main_ig:	rjmp	editor_main_1		

;------------------------------------------------------------------------------
;del char on cursor pos
;------------------------------------------------------------------------------
editor_dchar:	libmio_ecalc			;get memory position
		mov	tempreg2,XL		;copy x-coordinate
editor_dchar_1:	cpi	tempreg2,libmio_ecols-4	;at the end?
		breq	editor_dchar_2
		ldd	tempreg1,Y+1
		st	Y+,tempreg1
		inc	tempreg2
		rjmp	editor_dchar_1	
		
editor_dchar_2:	ldi	tempreg2,0x20			;space
		st	Y,tempreg2
		ret


;------------------------------------------------------------------------------
;check keys for Y/N 
;------------------------------------------------------------------------------
editor_yn:	libmio_waitkey
		cpi	tempreg1,'Y'
		breq	editor_yn1
		cpi	tempreg1,'y'
		breq	editor_yn1
		cpi	tempreg1,0xea		;ENTER
		breq	editor_yn1
		cpi	tempreg1,'N'
		breq	editor_yn2
		cpi	tempreg1,'n'
		breq	editor_yn2
		cpi	tempreg1,0xed		;ESC
		breq	editor_yn2
		rjmp	editor_yn
		
editor_yn1:	set				;set T-flag
		ret		
editor_yn2:	clt				;res T-flag
		ret		
		
;------------------------------------------------------------------------------
;check for unchanged-symbol
;------------------------------------------------------------------------------
editor_uc:	lds	XL,libmio_vram
		cpi	XL,editor_unchanged	;no change
		ret
		
		
;------------------------------------------------------------------------------
;clear status line 
;------------------------------------------------------------------------------
editor_cstat:	ldi	XL,0xcf
		rcall	editor_fstat
		libmio_thistext
		.db 2,2,"OK",0
		ret

;------------------------------------------------------------------------------
;fill status line with XL 
;------------------------------------------------------------------------------
editor_fstat:	push	ZL
		push	ZH
		ldi	ZL,LOW(libmio_vram+libmio_cols*2)	
		ldi	ZH,HIGH(libmio_vram+libmio_cols*2)	
		ldi	XH,libmio_cols
editor_fstat1:	st	Z+,XL
		dec	XH
		brne	editor_fstat1
		pop	ZH
		pop	ZL		
		ret

;-----------------------------------------------------------------------
; edit name
;-----------------------------------------------------------------------
editor_ren:	ldi	XH,0
		ldi	XL,9
		libmio_gotoxy
edit_ren_01:	libmio_cursor			;view cursor
		cpi	tempreg1,0xea		;enter
		brne	edit_ren_02
		ret
edit_ren_02:	cpi	tempreg1,0xe2		;cursor left
		brne	edit_ren_03
		mov	XL,libmio_cur_x
		cpi	XL,11
		breq	edit_ren_01
		dec	XL
		mov	libmio_cur_x,XL
		rjmp	edit_ren_01
edit_ren_03:	cpi	tempreg1,0xe3		;cursor right
		brne	edit_ren_04
		mov	XL,libmio_cur_x
		cpi	XL,24
		breq	edit_ren_01
		inc	XL
		mov	libmio_cur_x,XL
		rjmp	edit_ren_01
edit_ren_04:	cpi	tempreg1,0x20
		brcs	edit_ren_01			
		cpi	tempreg1,0x80
		brcc	edit_ren_01
		st	Y,tempreg1
		mov	XL,libmio_cur_x
		cpi	XL,24
		breq	edit_ren_05
		inc	XL
		mov	libmio_cur_x,XL
edit_ren_05:	rjmp	edit_ren_01

;------------------------------------------------------------------------------
;status messages
;------------------------------------------------------------------------------
errtab1:	.db "BREAK",0,"OVERFLOW",0,"DIVIDE /0",0,"SQR FROM <0"
errtab5:	.db 0,"CONSTANT TOO BIG",0,"WRONG EXPRESSION",0,"SYNTAX ERROR",0,"UNKNOWN KEYWORD",0
errtab9:	.db "WRONG FORMAT",0,"BAD LINENUMBER",0,"NEXT W/O FOR",0,"RETURN W/O CALL"
errtab13:	.db 0,"TOO MANY CALL",0,"TOO MANY FOR",0,"I2C ERROR",0,"UNKNOWN ERROR",0
errtab17:	.db "DFLASH ERROR",0,"OUT OF ARRAY",0,0